/*
 * Copyright 2017-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.testbase.knowledge.iterator.CollectionIterator;
import ru.ispras.testbase.knowledge.iterator.Iterator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

/**
 * The {@link GeneratorBoundaryValues} class generates boundary values related
 * to integer arithmetic for bit vectors of the specified size.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
final class GeneratorBoundaryValues extends DataGeneratorBase {
  public static final String NAME = "boundary";

  /** The cache of values for various bit vector sizes. */
  private static Map<Integer, Collection<NodeValue>> boundaryValues = new HashMap<>();

  public GeneratorBoundaryValues() {
    super(NAME);
  }

  @Override
  protected void generate(final DataGeneratorQuery query) {
    for (final DataGeneratorVariable variable : query.getUnknownVariables()) {
      if (variable.isType(DataTypeId.BIT_VECTOR)) {
        setValue(variable, newBoundaryIterator(variable.getSize()));
      } else {
        throw new IllegalArgumentException(String.format(
            "Variable %s has unsupported type %s.", variable.getName(), variable.getType()));
      }
    }
  }

  private static Iterator<NodeValue> newBoundaryIterator(final int valueSize) {
    Collection<NodeValue> values = boundaryValues.get(valueSize);

    if (null == values) {
      values = newNodeValueCollection(newBoundaryValues(valueSize));
      boundaryValues.put(valueSize, values);
    }

    return new CollectionIterator<>(values);
  }

  private static Collection<NodeValue> newNodeValueCollection(final Collection<BitVector> values) {
    final List<NodeValue> result = new ArrayList<>();
    for (final BitVector value : values) {
      result.add(NodeValue.newBitVector(value));
    }
    return result;
  }

  private static Collection<BitVector> newBoundaryValues(final int valueSize) {
    final Collection<BitVector> result = new HashSet<>();

    // 0x0
    result.add(BitVector.valueOf(0, valueSize));
    // 0x1
    result.add(BitVector.valueOf(1, valueSize));
    // 0x10..0
    BitVector temp = BitVector.valueOf(0, valueSize);
    temp.setBit((valueSize - 1), true);
    result.add(temp.copy());
    // 0xf..f
    temp.setAll();
    result.add(temp.copy());
    // 0x7f..f
    temp.setBit((valueSize - 1), false);
    result.add(temp.copy());
    // 0x10..01
    temp = BitVector.valueOf(1, valueSize);
    temp.setBit((valueSize - 1), true);
    result.add(temp.copy());

    BitVector temp2 = BitVector.valueOf(1, valueSize);
    int step = 2; // 2 bits
    while (valueSize / step > 0) {
      temp = temp2;
      temp2 = BitVectorMath.shl(temp2, step / 2);
      temp2 = BitVectorMath.add(temp, temp2);
      result.add(temp2);
      step = step * 2;
    }

    return result;
  }
}
